# -*- coding: utf-8 -*-
###
### $Release: 1.0.0 $
### Copyright (c) 2024-present Hyun-Gyu Kim (babyworm@gmail.com). MIT License.
### Original: copyright(c) 2007-2012 kuwata-lab.com all rights reserved.
###

import pytest
import re
import re as _re
import sys, os, traceback, time
import yaml

from testcase_helper import *
import tenjin
from tenjin.helpers import escape, to_str

PYPY   = hasattr(sys, 'pypy_version_info')
JYTHON = hasattr(sys, 'JYTHON_JAR')

def _errmsg(errmsg):
    if PYPY:
        if re.match(r"^name '.*' is not defined$", errmsg):
            return "global " + errmsg
    return errmsg

from tenjin.cli import Main, CommandOptionError, NoTextTemplate
import tenjin.cli as _cli_mod

def to_list(value):
    if isinstance(value, list):
        return value
    return [value]


import sys
python_version = sys.version.split(' ')[0]


INPUT = r"""<ul>
<?py for item in ['<a&b>', '["c",'+"'d']"]: ?>
  <li>#{item}
      ${item}</li>
<?py #end ?>
</ul>
"""
INPUT2 = """<ul>\r\n\
<?py for item in ['<a&b>', '["c",'+"'d']"]: ?>\r\n\
  <li>#{item}\r\n\
      ${item}</li>\r\n\
<?py #end ?>\r\n\
</ul>\r\n\
"""
INPUT3 = r"""<?py
#title = _context['title']
#items = _context.items
?>
<h1>#{title}</h1>
<ul>
<?py for item in items: ?>
  <li>#{item}</li>
<?py #endfor ?>
</ul>
"""

SOURCE = r"""_buf = []; _extend=_buf.extend;_to_str=to_str;_escape=escape; _extend(('''<ul>\n''', ));
for item in ['<a&b>', '["c",'+"'d']"]:
    _extend(('''  <li>''', _to_str(item), '''
      ''', _escape(_to_str(item)), '''</li>\n''', ));
#end
_extend(('''</ul>\n''', ));
print(''.join(_buf))
"""
SOURCE2 = """_buf = []; _extend=_buf.extend;_to_str=to_str;_escape=escape; _extend(('''<ul>\\r\\n''', ));\n\
for item in ['<a&b>', '["c",'+"'d']"]:\n\
    _extend(('''  <li>''', _to_str(item), '''\\r\n\
      ''', _escape(_to_str(item)), '''</li>\\r\\n''', ));\n\
#end\n\
_extend(('''</ul>\\r\\n''', ));\n\
print(''.join(_buf))
"""
SOURCE_N = r"""    1:  _buf = []; _extend=_buf.extend;_to_str=to_str;_escape=escape; _extend(('''<ul>\n''', ));
    2:  for item in ['<a&b>', '["c",'+"'d']"]:
    3:      _extend(('''  <li>''', _to_str(item), '''
    4:        ''', _escape(_to_str(item)), '''</li>\n''', ));
    5:  #end
    6:  _extend(('''</ul>\n''', ));
    7:  print(''.join(_buf))
"""

EXECUTED = r"""<ul>
  <li><a&b>
      &lt;a&amp;b&gt;</li>
  <li>["c",'d']
      [&quot;c&quot;,&#39;d&#39;]</li>
</ul>
"""
EXECUTED2 = """<ul>\r\n\
  <li><a&b>\r\n\
      &lt;a&amp;b&gt;</li>\r\n\
  <li>["c",'d']\r\n\
      [&quot;c&quot;,&#39;d&#39;]</li>\r\n\
</ul>\r\n\
"""
EXECUTED3 = r"""<h1>tenjin example</h1>
<ul>
  <li>aaa</li>
  <li>bbb</li>
  <li>ccc</li>
</ul>
"""

CONTEXT1 = r"""
title: tenjin example
items:
	- aaa
	- bbb
	- ccc
"""
CONTEXT2 = r"""
title = 'tenjin example'
items = ['aaa', 'bbb', 'ccc']
"""


class TestMain:

    def setup_method(self):
        pass

    def teardown_method(self):
        pass

    def _test(self):
        input     = getattr(self, 'input', '')
        source    = getattr(self, 'source', None)
        expected  = getattr(self, 'expected', None)
        exception = getattr(self, 'exception', None)
        errormsg  = getattr(self, 'errormsg', None)
        options   = getattr(self, 'options', '')
        filename  = getattr(self, 'filename', None)
        context_file = getattr(self, 'context_file', None)
        context_data = getattr(self, 'context_data', None)
        encoding  = getattr(self, 'encoding', None)
        #
        if python_version < '2.5':
            if expected:
                expected = expected.replace(': unexpected indent', ': invalid syntax')
        #
        if JYTHON:
            if self._testMethodName == 'test_lint5':
                expected = (
                    ".test.pyhtml:5:4: mismatched input 'else' expecting DEDENT\n"
                    "  5:     else\n"
                    "        ^\n"
                    )
        #
        if PYPY:
            if errormsg:
                errormsg = _errmsg(errormsg)
            if self._testMethodName == 'test_lint5':
                expected = (
                    ".test.pyhtml:5:4: invalid syntax\n"
                    "  5:     else\n"
                    "        ^\n"
                    )
        #
        if filename is not False:
            if filename is None:
                filename = '.test.pyhtml'
            #write_file(filename, input)
            for fname, s in zip(to_list(filename), to_list(input)):
                if encoding and isinstance(s, _unicode):
                    s = s.encode(encoding)
                write_file(fname, s)
        #
        if isinstance(options, list):
            argv = options
        elif isinstance(options, str):
            argv = [item for item in options.split(' ') if item]
        argv.insert(0, 'tenjin')
        if filename:
            #argv.append(filename)
            argv.extend(to_list(filename))
        #print "*** debug: argv=%s" % repr(argv)
        #
        if context_file:
            s = context_data
            if encoding and instance(s, _unicode):
                s = s.encode(encoding)
            write_file(context_file, s)
        #
        try:
            app = Main(argv)
            if exception:
                lst = [None]
                def f1():
                    try:
                        output = app.execute()
                    except Exception:
                        ex = sys.exc_info()[1]
                        lst[0] = ex
                        raise ex
                with pytest.raises(exception):
                    f1()
                if errormsg:
                    ex = lst[0]
                    assert str(ex) == errormsg
            else:
                output = app.execute()
                #print "*** expected=%s" % expected
                #print "*** output=%s" % output
                if python2:
                    if encoding and isinstance(output, unicode):
                        output = output.encode(encoding)
                assert output == expected
        finally:
            try:
                if filename:
                    #os.remove(filename)
                    for fname in to_list(filename):
                        os.unlink(fname)
                if context_file:
                    os.remove(context_file)
            except:
                pass

    code = TestCaseHelper.generate_testcode(__file__)
    exec(code)



    def test_help(self):  # -h, --help
        self.options  = "-h"
        self.input    = ""
        self.expected = Main(['tenjin']).usage('tenjin')
        self._test()
        #
        self.options  = "--help"
        self._test()

    def test_version(self):  # -v, --version
        self.options  = "-v"
        self.input    = ""
        self.expected = Main(['tenjin']).version() + "\n"
        self._test()
        self.options = '--version'
        self._test()

#    def test_help_and_version(self):  # -hVz
#        self.options  = "-hVc"
#        self.input    = "<?py foo() ?>"
#        app = Main(['tenjin'])
#        self.expected = app.version() + "\n" + app.usage('tenjin')
#        self._test()

    def test_render(self):  # (nothing), -a render
        self.options  = ""
        self.input    = INPUT
        self.expected = EXECUTED
        self._test()
        self.options  = "-a render"
        self._test()

    def test_source(self):  # -s, -a convert
        self.options  = "-s"
        self.input    = INPUT
        self.expected = SOURCE
        self._test()
        self.options = "-a convert"
        self._test()

    def test_source2(self):  # -s, -aconvert
        self.options  = "-s"
        n1 = len("<ul>\n")
        n2 = len("</ul>\n")
        self.input    = INPUT[n1:-n2]
        buf = SOURCE.splitlines(True)[1:-2]
        buf.insert(0, "_buf = []\n_extend=_buf.extend;_to_str=to_str;_escape=escape; \n")
        buf.append("print(''.join(_buf))\n")
        self.expected = ''.join(buf)
        self._test()
        self.options = "-aconvert"
        self._test()

    def test_source3(self):  # -sb, -baconvert
        self.options  = "-sb"
        self.input    = INPUT
        n1 = len("_buf = []; ")
        n2 = len("print(''.join(_buf))\n")
        self.expected = SOURCE[n1:-n2]
        self._test()
        self.options = "-baconvert"
        self._test()

    def test_number1(self):   # -sN
        self.options  = "-sN"
        self.input    = INPUT
        self.expected = SOURCE_N
        self._test()

    def test_number2(self):   # -sbN
        self.options  = "-sbN"
        self.input    = INPUT
        self.expected = re.sub(r'\n    7:.*?\n$', "\n", SOURCE_N).replace('_buf = []; ', '')
        self._test()

    def test_cache1(self):   # -a cache
        self.options  = "-a cache"
        self.input    = (
            '<?py #@ARGS title, items ?>\n'
            '<h1>${title}</h1>\n'
            '<ul>\n'
            '<?py for item in items: ?>\n'
            '  <li>${item}</li>\n'
            '<?py #endfor ?>\n'
            '</ul>\n'
            )
        self.expected = ''
        script = (
            "title = _context.get('title'); items = _context.get('items'); \n"
            "_extend=_buf.extend;_to_str=to_str;_escape=escape; _extend(('''<h1>''', _escape(_to_str(title)), '''</h1>\n"
            "<ul>\\n''', ));\n"
            "for item in items:\n"
            "    _extend(('''  <li>''', _escape(_to_str(item)), '''</li>\\n''', ));\n"
            "#endfor\n"
            "_extend(('''</ul>\\n''', ));\n"
            )
        expected_cache = r"""
timestamp: %(timestamp)s
args: title, items

title = _context.get('title'); items = _context.get('items'); 
_extend=_buf.extend;_to_str=to_str;_escape=escape; _extend(('''<h1>''', _escape(_to_str(title)), '''</h1>
<ul>\n''', ));
for item in items:
    _extend(('''  <li>''', _escape(_to_str(item)), '''</li>\n''', ));
#endfor
_extend(('''</ul>\n''', ));
"""[1:]  % {'timestamp': '%s.0' % int(time.time())}
        self.filename = 'test_cache1.pyhtml'
        cachename = self.filename + '.cache'
        try:
            self._test()
            assert os.path.exists(cachename)
            #if not JYTHON:
            #    import marshal
            #    dct = marshal.load(open(cachename, 'rb'))
            #    ok (dct.get('args')) == ['title', 'items']
            #    if   python2:  expected = "<type 'code'>"
            #    elif python3:  expected = "<class 'code'>"
            #    ok (str(type(dct.get('bytecode')))) == expected
            #    ok (dct.get('script')) == script
            f = open(cachename); cached = f.read(); f.close()
            # Normalize timestamp precision for Python 3.12+
            cached = re.sub(r'timestamp: (\d+)\.\d+', r'timestamp: \1.0', cached)
            assert cached == expected_cache
        finally:
            if os.path.exists(cachename):
                os.unlink(cachename)

    input_for_retrieve = (
        '<div>\n'
        '<?py if list: ?>\n'
        '  <table>\n'
        '    <thead>\n'
        '      <tr>\n'
        '        <th>#</th><th>item</th>\n'
        '      </tr>\n'
        '    </thead>\n'
        '    <?py i = 0 ?>\n'
        '    <?py for item in list: ?>\n'
        '\t<?py i += 1 ?>\n'
        '    <tbody>\n'
        '      <tr bgcolor="#{i % 2 and "#FFCCCC" or "#CCCCFF"}">\n'
        '\t<td>${i}</td>\n'
        '        <td>${item}</td>\n'
        '      </tr>\n'
        '    </tbody>\n'
        '    <?py #end ?>\n'
        '  </table>\n'
        '<?py #end ?>'
        '</div>\n'
        )
    expected_for_retrieve = '\n'.join((
        '_buf = []; _extend=_buf.extend;_to_str=to_str;_escape=escape; ',
        'if list:',
        '',
        '',
        '',
        '',
        '',
        '',
        '    i = 0',
        '    for item in list:',
        '        i += 1',
        '',
        '        _to_str(i % 2 and "#FFCCCC" or "#CCCCFF"); ',
        '        _escape(_to_str(i)); ',
        '        _escape(_to_str(item)); ',
        '',
        '',
        '    #end',
        '',
        '#end',
        '',
        'print(\'\'.join(_buf))',
        ''))

    def test_retrieve1(self):  # -S, -a retrieve
        self.input    = self.input_for_retrieve
        self.expected = self.expected_for_retrieve
        self.options = '-S'
        self._test()
        self.options = '-a retrieve'
        #self._test()

    def test_retrieve2(self):  # -SU, -SNU
        expected = '\n'.join((
            '_buf = []; _extend=_buf.extend;_to_str=to_str;_escape=escape; ',
            'if list:',
            '',
            '    i = 0',
            '    for item in list:',
            '        i += 1',
            '',
            '        _to_str(i % 2 and "#FFCCCC" or "#CCCCFF"); ',
            '        _escape(_to_str(i)); ',
            '        _escape(_to_str(item)); ',
            '',
            '    #end',
            '',
            '#end',
            '',
            'print(\'\'.join(_buf))',
            ''))
        self.input = self.input_for_retrieve
        self.expected = expected
        self.options = '-SU'
        self._test()
        #
        expected = '\n'.join((
            '    1:  _buf = []; _extend=_buf.extend;_to_str=to_str;_escape=escape; ',
            '    2:  if list:',
            '',
            '    9:      i = 0',
            '   10:      for item in list:',
            '   11:          i += 1',
            '',
            '   13:          _to_str(i % 2 and "#FFCCCC" or "#CCCCFF"); ',
            '   14:          _escape(_to_str(i)); ',
            '   15:          _escape(_to_str(item)); ',
            '',
            '   18:      #end',
            '',
            '   20:  #end',
            '',
            '   22:  print(\'\'.join(_buf))',
            ''))
        self.expected = expected
        self.options = '-SNU'
        self._test()

    def test_retrieve3(self):  # -SC, -SNC
        expected = '\n'.join((
            '_buf = []; _extend=_buf.extend;_to_str=to_str;_escape=escape; ',
            'if list:',
            '    i = 0',
            '    for item in list:',
            '        i += 1',
            '        _to_str(i % 2 and "#FFCCCC" or "#CCCCFF"); ',
            '        _escape(_to_str(i)); ',
            '        _escape(_to_str(item)); ',
            '    #end',
            '#end',
            'print(\'\'.join(_buf))',
            ''))
        self.input = self.input_for_retrieve
        self.expected = expected
        self.options = '-SC'
        self._test()
        #
        expected = '\n'.join((
            '    1:  _buf = []; _extend=_buf.extend;_to_str=to_str;_escape=escape; ',
            '    2:  if list:',
            '    9:      i = 0',
            '   10:      for item in list:',
            '   11:          i += 1',
            '   13:          _to_str(i % 2 and "#FFCCCC" or "#CCCCFF"); ',
            '   14:          _escape(_to_str(i)); ',
            '   15:          _escape(_to_str(item)); ',
            '   18:      #end',
            '   20:  #end',
            '   22:  print(\'\'.join(_buf))',
            ''))
        self.expected = expected
        self.options = '-SNC'
        self._test()

    def test_statements(self):  # -X, -a statements
        expected = '\n'.join((
            '_buf = []; _extend=_buf.extend;_to_str=to_str;_escape=escape; ',
            'if list:',
            '',
            '',
            '',
            '',
            '',
            '',
            '    i = 0',
            '    for item in list:',
            '        i += 1',
            '',
            '',
            '',
            '',
            '',
            '',
            '    #end',
            '',
            '#end',
            '',
            'print(\'\'.join(_buf))',
            ''))
        self.input = self.input_for_retrieve
        self.expected = expected
        self.options = '-X'
        self._test()
        self.options = '-a statements'
        self._test()

    def test_dump(self):  # -d, -a dump
        if JYTHON:
            return
        ## create cache file
        #filename = '_test_dump.pyhtml'
        #cachename = filename + '.cache'
        #self.filename = filename
        #self.input = INPUT
        #self.expected = EXECUTED
        #self.options = '-a render --cache=true'
        #self._test()
        #ok (cachename).exists()
        ## create marshal cache file
        filename = '_test_dump.pyhtml'
        cachename = filename + '.cache'
        f = open(filename, 'w'); f.write(INPUT); f.close()
        e = tenjin.Engine(cache=tenjin.MarshalCacheStorage())
        e.get_template(filename)
        assert os.path.exists(cachename)
        # dump test
        try:
            self.filename = False
            self.input    = False
            self.expected = SOURCE[len('_buf = []; '):-len("print(''.join(_buf))\n")]
            #self.options = '-d %s' % cachename
            #self._test()
            self.options = '-a dump %s' % cachename
            self._test()
        finally:
            os.unlink(cachename)

    def test_indent(self):  # -i2
        self.options  = "-si2"
        self.input    = INPUT
        pat = re.compile(r'^    _extend', re.M)
        self.expected = pat.sub(r'  _extend',  SOURCE)
        self._test()

    def test_quiet(self):  # -q, -qasyntax
        self.options  = "-z"
        input = INPUT
        self.input    = [input, input, input]
        basename = ".test_quiet%d.pyhtml"
        self.filename = [basename % i for i in range(0,3)]
        self.expected = ''.join([(basename+" - ok.\n") % i for i in range(0,3)])
        self._test()
        #
        self.options  = "-zq"
        self.expected = ""
        self._test()
        self.options = "-qasyntax"
        self._test()

    def test_invalid_options(self):  # -Y, -i, -f, -c, -i foo
        self.input    = INPUT
        self.expected = ""
        self.exception = CommandOptionError
        #
        self.options  = "-hY"
        self.errormsg = "-Y: unknown option."
        self._test()
        #
        self.options  = "-i"
        self.filename = False
        #self.errormsg = "-i: indent width required."
        self.errormsg = "-i: argument required."
        self._test()
        #
        self.options  = "-f"
        self.filename = False
        #self.errormsg = "-f: context data filename required."
        self.errormsg = "-f: argument required."
        self._test()
        #
        self.options  = "-c"
        self.filename = False
        #self.errormsg = "-c: context data string required."
        self.errormsg = "-c: argument required."
        self._test()
        #
        self.options  = "-i foo"
        self.errormsg = "-i: integer value required."
        self._test()
        #

    def test_newline(self):
        self.options  = "-s"
        self.input    = INPUT2
        self.expected = SOURCE2
        self._test()
        self.options  = ""
        self.expected = EXECUTED2
        self._test()

    def test_datafile_yaml(self): # -f datafile.yaml
        context_filename = 'test.datafile.yaml'
        self.options  = "-f " + context_filename
        self.input    = INPUT3
        self.expected = EXECUTED3
        self.context_file = context_filename
        self.context_data = CONTEXT1
        self._test()

    def test_datafile_py(self): # -f datafile.py
        context_filename = 'test.datafile.py'
        self.options  = "-f " + context_filename
        self.input    = INPUT3
        self.expected = EXECUTED3
        self.context_file = context_filename
        self.context_data = CONTEXT2
        self._test()

    def test_datafile_error(self):  # -f file.txt, not-a-mapping context data
        context_filename = 'test.datafile.txt'
        self.options = "-f " + context_filename
        self.exception = CommandOptionError
        self.errormsg = "-f %s: file not found." % context_filename
        self._test()
        #
        self.context_file = context_filename
        self.context_data = "- foo\n- bar\n -baz"
        self.errormsg = "-f %s: unknown file type ('*.yaml' or '*.py' expected)." % context_filename
        self._test()
        #
        context_filename = 'test.datafile.yaml'
        self.options = "-f " + context_filename
        self.errormsg = "%s: not a mapping (dictionary)." % context_filename
        self.context_file = context_filename
        self._test()

    def test_context_yaml(self):  # -c yamlstr
        self.options = ['-c', '{title: tenjin example, items: [aaa, bbb, ccc]}']
        self.input    = INPUT3
        self.expected = EXECUTED3
        self._test()

    def test_context_py(self):  # -c python-code
        self.options = ['-c', 'title="tenjin example";  items=["aaa", "bbb", "ccc"]']
        self.input    = INPUT3
        self.expected = EXECUTED3
        self._test()

    def test_untabify(self):  # -T
        context_filename = 'test.datafile.yaml'
        self.options  = "-Tf " + context_filename
        self.input    = INPUT3
        self.expected = EXECUTED3
        self.context_file = context_filename
        self.context_data = CONTEXT1
        self.exception = yaml.parser.ScannerError
        self._test()

    def test_modules(self):  # -r modules
        #self.options  = "--escapefunc=cgi.escape"
        #self.input    = INPUT
        #self.expected = EXECUTED.replace('&quot;', '"')
        #self.exception = NameError
        #self.errormsg = "name 'cgi' is not defined"
        #self._test()
        ##
        #self.options  = "-r cgi,os,sys --escapefunc=cgi.escape"
        #self.input    = INPUT
        #self.expected = EXECUTED.replace('&quot;', '"')
        #self.exception = None
        #self.errormsg = None
        #self._test()
        #
        self.input    = "Hello #{html.escape('Haru&Kyon')}!"
        self.expected = "Hello Haru&amp;Kyon!"
        #
        _cli_mod.__dict__.pop('html', None)
        self.exception = NameError
        self.errormsg = "name 'html' is not defined"
        self._test()
        #
        self.options  = "-r html,os,sys"
        self.exception = None
        self.errormsg = None
        self._test()

    def test_modules_err(self):  # -r hogeratta
        self.options = '-r hogeratta'
        self.exception = CommandOptionError
        self.errormsg = '-r hogeratta: module not found.'
        self._test()

    def test_escapefunc(self):  # --escapefunc=html.escape
        self.options  = "-s --escapefunc=html.escape"
        self.input    = INPUT
        self.expected = SOURCE.replace('=escape', '=html.escape')
        self._test()

    def test_tostrfunc(self):  # --tostrfunc=str
        self.options  = "-s --tostrfunc=str"
        self.input    = INPUT
        self.expected = SOURCE.replace('=to_str', '=str')
        self._test()

    def test_preamble(self):  # --preamble --postamble
        self.options  = ["-s", "--preamble=_buf=list()", "--postamble=return ''.join(_buf)"]
        self.input    = INPUT
        self.expected = re.sub(r'print\((.*?)\)', r'return \1', SOURCE).replace("_buf = []; ", "_buf=list(); ")
        self._test()

    def test_xencoding1(self):  # --encoding=encoding
        if python2:
            self.input = """\
<?py items=['foo',u'bar',u'日本語'] ?>
ようこそ
<?py for item in items: ?>
* 「${item}」
<?py #end ?>
"""
        elif python3:
            self.input = """\
<?py items=['foo','bar','日本語'] ?>
ようこそ
<?py for item in items: ?>
* 「${item}」
<?py #end ?>
"""
        self.expected = """\
ようこそ
* 「foo」
* 「bar」
* 「日本語」
"""
        #
        self.encoding = 'utf-8'
        self.options  = "--encoding=%s" % self.encoding
        self._test()

    def test_xencoding2(self):  # -k encoding
        if python2:
            self.input = """\
<?py items=['foo',u'bar','日本語'] ?>
ようこそ
<?py for item in items: ?>
* 「${item}」
<?py #end ?>
"""
        elif python3:
            self.input = """\
<?py items=['foo',b'bar','日本語'] ?>
ようこそ
<?py for item in items: ?>
* 「${item}」
<?py #end ?>
"""
        self.expected = """\
ようこそ
* 「foo」
* 「bar」
* 「日本語」
"""
        #
        self.encoding = 'utf-8'
        self.options  = "-k utf-8"
        tostr_func = to_str
        try:
            self._test()
        finally:
            _cli_mod.__dict__['to_str'] = tostr_func

    def test_template_path(self):  # --path
        layout = r'''<html>
  <body>
#{_content}
<?py include(':footer') ?>
  </body>
</html>
'''
        body = r'''<ul>
<?py for item in 'ABC': ?>
  <li>${item}</li>
<?py #endfor ?>
</ul>
'''
        footer = r'''<hr />
<a href="mailto:webmaster@localhost">webmaser</a>
'''
        expected = r'''<html>
  <body>
<ul>
  <li>A</li>
  <li>B</li>
  <li>C</li>
</ul>

<hr />
<a href="mailto:webmaster@localhost">webmaser</a>
  </body>
</html>
'''
        try:
            os.mkdir("tmpl9")
            os.mkdir("tmpl9/user")
            write_file("tmpl9/layout.pyhtml", layout)
            write_file("tmpl9/body.pyhtml", '')
            write_file("tmpl9/footer.pyhtml", '')
            write_file("tmpl9/user/body.pyhtml", body)
            write_file("tmpl9/user/footer.pyhtml", footer)
            self.options  = "--path=.,tmpl9/user,tmpl9 --postfix=.pyhtml --layout=:layout"
            self.input = "<?py include(':body') ?>"
            self.expected = expected
            self._test()
        finally:
            from glob import glob
            for f in glob('tmpl9/user/*.pyhtml'): os.unlink(f)
            for f in glob('tmpl9/*.pyhtml'): os.unlink(f)
            for d in ['tmpl9/user', 'tmpl9']: os.rmdir(d)

    def test_preprocess1(self):  # -P, -a preprocess, --preprocess
        input = '''\
<?PY states = { "CA": "California", ?>
<?PY            "NY": "New York", ?>
<?PY            "FL": "Florida", } ?>
<?PY # ?>
<?py chk = { params['state']: ' checked="checked"' } ?>
<?PY codes = list(states.keys()) ?>
<?PY codes.sort() ?>
<select name="state">
  <option value="">-</option>
<?PY for code in codes: ?>
  <option value="#{{code}}"#{chk.get(#{{repr(code)}}, '')}>${{states[code]}}</option>
<?PY #endfor ?>
</select>
'''
        script = '''\
<?py chk = { params['state']: ' checked="checked"' } ?>
<select name="state">
  <option value="">-</option>
  <option value="CA"#{chk.get('CA', '')}>California</option>
  <option value="FL"#{chk.get('FL', '')}>Florida</option>
  <option value="NY"#{chk.get('NY', '')}>New York</option>
</select>
'''
        expected = '''\
<select name="state">
  <option value="">-</option>
  <option value="CA" checked="checked">California</option>
  <option value="FL">Florida</option>
  <option value="NY">New York</option>
</select>
'''
        try:
            self.options  = "-P" # "-P --prefix=prep_ --postfix=.pyhtml"
            self.input = input
            self.expected = script
            self._test()
            self.options = "-a preprocess"
            self._test()
            #
            self.options  = ["--preprocess=true", "-c", "{params: {state: CA}}"]
            self.input = input
            self.expected = expected
            self._test()
        finally:
            pass

    def test_safe(self):  # --safe
        _backup = tenjin.Engine.templateclass
        try:
            self.options  = "-s --safe"
            self.input    = INPUT
            self.input    = re.sub(r'#{(.*?)}', r'{==\1==}', self.input)
            self.input    = re.sub(r'\${(.*?)}', r'{=\1=}', self.input)
            self.expected = SOURCE.replace('=escape', '=to_escaped')
            self.expected = re.sub(r'_escape\(_to_str\((.*?)\)\)', r'_escape(\1)', self.expected)
            self._test()
        finally:
            tenjin.Engine.templateclass = _backup

    def test_pp(self):  # --pp=name,name2,...
        input = r"""
<div id="placeholder">
  <!-- #JS: render_table(items) -->
  <table>
    <?js for (var i = 0; i < items.length; i++) { ?>
    <tr>
      <td>${item}</td>
    </tr>
    <?js } ?>
  </table>
  <!-- #/JS -->
</div>
<script>#{tenjin.JS_FUNC}</script>
"""[1:]
        output = r"""
<div id="placeholder">
<script>function render_table(items){var _buf='';
_buf+='<table>\n';
 for (var i = 0; i < items.length; i++) {
_buf+='<tr>\n\
<td>'+_E(item)+'</td>\n\
</tr>\n';
 }
_buf+='</table>\n';
return _buf;};</script>
</div>
<script>function _S(x){return x==null?'':x;}
function _E(x){return x==null?'':typeof(x)!=='string'?x:x.replace(/[&<>"']/g,_EF);}
var _ET={'&':"&amp;",'<':"&lt;",'>':"&gt;",'"':"&quot;","'":"&#039;"};
function _EF(c){return _ET[c];};</script>
"""[1:]
        try:
            self.options  = "--pp=Trim,PrefixedLine,JavaScript"
            self.input = input
            self.expected = output
            self._test()
        finally:
            pass



# -*- coding: utf-8 -*-
import tenjin
from tenjin.helpers import *
import re

##
## message catalog to translate message
##
MESSAGE_CATALOG = {
    'en': { 'Hello': 'Hello',
            'Good bye': 'Good bye',
          },
    'fr': { 'Hello': 'Bonjour',
            'Good bye': 'Au revoir',
          },
}

##
## create translation function and return it.
## ex.
##    _ = create_m17n_func('fr')
##    print _('Hello')   #=> 'Bonjour'
##
def create_m17n_func(lang):
    dct = MESSAGE_CATALOG.get(lang)
    if not dct:
        raise ValueError("%s: unknown lang." % lang)
    def _(message_key):
        return dct.get(message_key)
    return _
    # or return dct.get

##
## test program
##
if __name__ == '__main__':

    ## render html for English
    engine_en = tenjin.Engine(preprocess=True, lang='en')
    context = { 'username': 'World' }
    context['_'] = create_m17n_func('en')
    html = engine_en.render('m17n.pyhtml', context)
    print("--- lang: en ---")
    print(html)

    ## render html for French
    engine_fr = tenjin.Engine(preprocess=True, lang='fr')
    context = { 'username': 'World' }
    context['_'] = create_m17n_func('fr')
    html = engine_fr.render('m17n.pyhtml', context)
    print("--- lang: fr ---")
    print(html)



value = 'My Great Example'

## create engine object with preprocessing enabled
import tenjin
from tenjin.helpers import *
engine = tenjin.Engine(path=['views'], preprocess=True)

## print Python script code
print("------ converted script ------")
print(engine.get_template('pp-example1.pyhtml').script)

## render html
html = engine.render('pp-example1.pyhtml', {})
print("------ rendered html ------")
print(html)



###
### $Release: 1.0.0 $
### Copyright (c) 2024-present Hyun-Gyu Kim (babyworm@gmail.com). MIT License.
### Original: copyright(c) 2007-2012 kuwata-lab.com all rights reserved.
###

import pytest
import sys, os, time
import tenjin

from test_engine import _with_dummy_files


class TestFileSystemLoader:

    def setup_method(self):
        self.loader = tenjin.FileSystemLoader()
        self.dirs = ['_views/blog', '_views']

    @_with_dummy_files
    def test_find(self):
        # if dirs provided then search template file from it.
        assert self.loader.find('index.pyhtml', self.dirs) == '_views/blog/index.pyhtml'
        assert self.loader.find('layout.pyhtml', self.dirs) == '_views/layout.pyhtml'
        # if dirs not provided then just return filename if file exists.
        assert self.loader.find('_views/index.pyhtml') == '_views/index.pyhtml'
        # if file not found then return None.
        assert self.loader.find('index2.pyhtml', self.dirs) == None
        assert self.loader.find('index2.pyhtml') == None

    @_with_dummy_files
    def test_abspath(self):
        # return full-path of filepath
        ret = self.loader.abspath('_views/blog/index.pyhtml')
        assert ret == os.path.join(os.getcwd(), '_views/blog/index.pyhtml')

    @_with_dummy_files
    def test_timestamp(self):
        # return mtime of file
        ts = float(int(time.time())) - 3.0
        os.utime('_views/blog/index.pyhtml', (ts, ts))
        ret = self.loader.timestamp('_views/blog/index.pyhtml')
        assert ret == ts

    @_with_dummy_files
    def test_read(self):
        # if file exists, return file content and mtime
        ts = float(int(time.time())) - 1.0
        os.utime('_views/layout.pyhtml', (ts, ts))
        ret = self.loader.load('_views/layout.pyhtml')
        assert ret == ("<div>#{_content}</div>", ts)
        # if file not exist, return None
        ret = self.loader.load('_views/layout2.pyhtml')
        assert ret == None



languages = [
  ('en', 'Engilish'),
  ('fr', 'French'),
  ('de', 'German'),
  ('es', 'Spanish'),
  ('ch', 'Chinese'),
  ('ja', 'Japanese'),
]

import urllib
try:
    from urllib import unquote
except:
    from urllib.parse import unquote

def link_to(label, action=None, id=None):
    buf = ['/app']
    if action: buf.append(action)
    if id: buf.append(id)
    return '<a href="%s">%s</a>' % (unquote('/'.join(buf)), label)



###
### $Release: 1.0.0 $
### Copyright (c) 2024-present Hyun-Gyu Kim (babyworm@gmail.com). MIT License.
### Original: copyright(c) 2007-2012 kuwata-lab.com all rights reserved.
###

#import unittest
import os, sys, difflib, re
import yaml

__all__ = ['TestCaseHelper', 'read_file', 'write_file',
           'remove_unmatched_test_methods',
           'python3', 'python2', '_unicode', '_bytes']

python3 = sys.version_info[0] == 3
python2 = sys.version_info[0] == 2
if python2:
    _unicode = unicode
    _bytes   = str
    def _is_str(val):
        return isinstance(val, (str, unicode))
elif python3:
    _unicode = str
    _bytes   = bytes
    def _is_str(val):
        return isinstance(val, (str, bytes))

JYTHON = hasattr(sys, 'JYTHON_JAR')


def read_file(filename, mode='rb'):
    f = None
    try:
        f = open(filename, mode)
        return f.read()
    finally:
        if f: f.close()

def write_file(filename, content, mode='wb'):
    if isinstance(content, _unicode):
        content = content.encode('utf-8')
    f = None
    try:
        f = open(filename, mode)
        f.write(content)
    finally:
        if f: f.close()

def remove_unmatched_test_methods(testcase_class, pattern=None):
    pattern = pattern or os.environ.get('TEST')
    if not pattern: return
    rexp = re.compile(pattern)
    for m in dir(testcase_class):
        if m.startswith('test_'):
            name = m[len('test_'):]
            if not rexp.search(name):
                delattr(testcase_class, m)


class TestCaseHelper:

    @classmethod
    def remove_tests_except(cls, test_name):
        if not test_name: return
        if not test_name.startswith('test_'):
            test_name = 'test_' + test_name
        for name in dir(cls):
            if name.startswith('test_') and name != test_name:
                delattr(cls, name)

    def _testname(self):
        try:
            return self._TestCase__testMethodName
        except AttributeError:
            return self._testMethodName

    def assertTextEqual(self, text1, text2, encoding=None):
        if text1 == text2:
            self.assertEqual(text1, text2)
        else:
            file1, file2 = '.tmp.file1', '.tmp.file2'
            if encoding:
                if isinstance(text1, _unicode):
                    text1 = text1.encode(encoding or 'utf-8')   # unicode to binary
                if isinstance(text2, _unicode):
                    text2 = text2.encode(encoding or 'utf-8')   # unicode to binary
            write_file(file1, text1)
            write_file(file2, text2)
            f = os.popen("diff -u %s %s" % (file1, file2))
            output = f.read()
            f.close()
            os.unlink(file1)
            os.unlink(file2)
            mesg = re.sub(r'.*?\n', '', output, 2)
            self.assertEqual(text1, text2, mesg)

    def assertRaise(self, exc_class, callable_, *args, **kwargs):
        try:
            callable_(*args, **kwargs)
        except exc_class:
            ex = sys.exc_info()[1]
            return ex
        else:
            exc_class_name = getattr(exc_class, '__name__', None) or str(exc_class)
            self.fail("%s not raised" % exc_class_name)

    def assertNotRaise(self, callable_, *args, **kwargs):
        try:
            return callable_(*args, **kwargs)
        except Exception:
            ex = sys.exc_info()[1]
            self.fail("unexpected exception raised: " + repr(ex))

    def assertExists(self, filename):
        if not os.path.exists(filename):
            self.fail("file not exist: " + repr(filename))

    def assertNotExist(self, filename):
        if os.path.exists(filename):
            self.fail("file exists: " + repr(filename))

    def assertFileExists(self, filename, identifier=None):
        msg = "File %s expected but nof found." % filename
        if identifier: msg = "[%s] %s" % (identifier, msg)
        self.assertTrue(os.path.isfile(filename), msg)

    def assertFileNotExist(self, filename, identifier=None):
        msg = "File %s expected not to be there but exists." % filename
        if identifier: msg = "[%s] %s" % (identifier, msg)
        self.assertFalse(os.path.isfile(filename), msg)

    def assertDirExists(self, dirname, identifier=None):
        msg = "Directory %s expected but nof found." % dirname
        if identifier: msg = "[%s] %s" % (identifier, msg)
        self.assertTrue(os.path.isdir(dirname), msg)

    def assertDirNotExist(self, dirname, identifier=None):
        msg = "Directory %s expected not to be there but exists." % dirname
        if identifier: msg = "[%s] %s" % (identifier, msg)
        self.assertFalse(os.path.isdir(dirname), msg)

    def assertFileNewerThan(self, filename, arg, identifier=None):
        msg = "File %s expected to be newer thant %s." % (repr(filename), repr(arg))
        if identifier: msg = "[%s] %s" % (identifier, msg)
        mtime1 = os.path.getmtime(filename)
        mtime2 = _is_str(arg) and os.path.getmtime(arg) or arg
        self.assertTrue(mtime1 > mtime2, msg)

    def assertSameTimestampWith(self, filename, arg, identifier=None):
        msg = "File %s expected to be the same timestampt with %s." % (repr(filename), repr(arg))
        if identifier: msg = "[%s] %s" % (identifier, msg)
        mtime1 = os.path.getmtime(filename)
        mtime2 = _is_str(arg) and os.path.getmtime(arg) or arg
        self.assertTrue(mtime1 == mtime2, msg)

    def asertFileOlderThan(self, filename, arg, identifier=None):
        msg = "File %s expected to be older thant %s." % (repr(filename), repr(arg))
        if identifier: msg = "[%s] %s" % (identifier, msg)
        mtime1 = os.path.getmtime(filename)
        mtime2 = _is_str(arg) and os.path.getmtime(arg) or arg
        self.assertTrue(mtime1 < mtime2, msg)

    ###

    def load_testdata(filename, untabify=True):
        i = filename.rfind('.')
        if JYTHON:
            if filename[i-3:i] == '$py':
                i -= 3
        if filename[i:] != '.yaml' and filename[i:] != '.yml':
            filename = filename[:i] + '.yaml'
        input = read_file(filename)
        if untabify:
            input = input.expandtabs()
        ydoc = yaml.safe_load(input)
        return ydoc
    load_testdata = staticmethod(load_testdata)

    def generate_testcode(filename, untabify=True, testmethod='_test', lang='python'):
        doclist = TestCaseHelper.load_testdata(filename, untabify)
        table = {}
        buf = []
        a = buf.append
        for doc in doclist:
            name = doc.get('name')
            if name is None:  raise Exception("'name:' is required.")
            if name in table: raise Exception("'name: %s' is duplicated." % name)
            table[name] = doc
            a(                             'def test_%s(self):' % name)
            for k, v in doc.items():
                if k[-1] == '*':
                    k = k[:-1]
                    v = v.get(lang)
                if k == 'exception':     a('    self.%s = %s'        % (k, v))
                elif isinstance(v, str): a('    self.%s = r"""%s"""' % (k, v))
                else:                    a('    self.%s = %s'        % (k, repr(v)))
            a(                             '    self.%s()' % testmethod)
            a(                             '#')
        a('')
        code = "\n".join(buf)
        return code
    generate_testcode = staticmethod(generate_testcode)


TestCaseHelper.assertRaises2 = TestCaseHelper.assertRaise



import os, tenjin
from tenjin.helpers import *

## create key-value store object
if not os.path.isdir('cache.d'): os.mkdir('cache.d')
kv_store = tenjin.FileBaseStore('cache.d')      # file based

## set key-value store into tenjin.helpers.fagment_cache object
tenjin.helpers.fragment_cache.store = kv_store

## context data
## (it is strongly recommended to create function object
##  to provide pull-style context data)
def get_items():   # called only when cache is expired
    return ['AAA', 'BBB', 'CCC']
context = {'get_items': get_items}

## render html
engine = tenjin.Engine(path=['views'])
html = engine.render('items.pyhtml', context)
print(html)



###
### $Release: 1.0.0 $
### Copyright (c) 2024-present Hyun-Gyu Kim (babyworm@gmail.com). MIT License.
### Original: copyright(c) 2007-2012 kuwata-lab.com all rights reserved.
###

import pytest
import re
import sys, os

from testcase_helper import *
import tenjin
from tenjin.helpers import *

PYPY   = hasattr(sys, 'pypy_version_info')
JYTHON = hasattr(sys, 'JYTHON_JAR')

def _errmsg(errmsg):
    if PYPY:
        if re.match(r"^name '.*' is not defined$", errmsg):
            return "global " + errmsg
    return errmsg


class TestTemplate:

    code = TestCaseHelper.generate_testcode(__file__)
    exec(code)


    def setup_method(self):
        pass

    def teardown_method(self):
        pass

    def _test(self):
        input     = getattr(self, 'input', None)
        source    = getattr(self, 'source', None)
        expected  = getattr(self, 'expected', None)
        exception = getattr(self, 'exception', None)
        errormsg  = getattr(self, 'errormsg', None)
        options   = getattr(self, 'options', {})
        filename  = getattr(self, 'filename', None)
        context   = getattr(self, 'context', {})
        testopts  = getattr(self, 'testopts', None)
        disabled  = getattr(self, 'disabled', None)
        templateclass = getattr(self, 'templateclass', None)
        encoding  = None
        #
        if disabled:
            return
        #
        if testopts:
            if 'crchar' in testopts:
                ch = testopts['crchar']
                if input:     input    = input.replace(ch, "\r")
                if source:    source   = source.replace(ch, "\r")
                if expected:  expected = expected.replace(ch, "\r")
            if testopts.get('escapefunc') == '_html_module.escape':
                #import cgi
                #context['escape'] = _html_module.escape
                pass
            if testopts.get('tostrfunc') == 'str':
                #context['to_str'] = str
                pass
            if 'encoding' in testopts:
                encoding = testopts.get('encoding')
            if 'templateclass' in testopts:
                templateclass = testopts.get('templateclass')
        #
        if python3:
            input  = input.replace('urllib.quote', 'urllib.parse.quote')
            source = source.replace('urllib.quote', 'urllib.parse.quote')
            if encoding:
                if source:
                    source = source.replace("u'''", "'''").replace("u'", "'")
                    input  = input.replace("u'", "'")
        #
        if JYTHON:
            if self._testMethodName == 'test_syntaxerr1':
                errormsg = r"mismatched input '\n' expecting COLON"
        #
        if PYPY:
            if errormsg:
                errormsg = _errmsg(errormsg)
        #
        if exception:
            with pytest.raises(exception) as exc_info:
                template = tenjin.Template(**options)
                template.convert(input, filename)
                template.render(context)
            ex = exc_info.value
            assert ex.__class__ == exception
            if errormsg:
                ## SyntaxError has 'msg' attribute instead of 'message'. Why?
                assert ex.args[0] == errormsg
            if filename:
                assert ex.filename == filename
        else:
            if templateclass:
                templateclass = eval(templateclass)
                template = templateclass(**options)
            else:
                template = tenjin.Template(**options)
            script = template.convert(input, filename)
            assert script == source         # encoding=encoding
            if expected:
                output = template.render(context)
                assert output == expected   # encoding=encoding



#    def test_render1(self):   # Tenjin#render(context) == Tenjin#render(**context)
#        input = """<ul>
#<?py for item in items: ?>
#<li>#{item}</li>
#<?py #endfor ?>
#</ul>
#"""
#        template = tenjin.Template()
#        template.convert(input)
#        items = ['foo', 'bar', 'baz']
#        context = {'items': items}
#        output1 = template.render(context)
#        output2 = template.render(items=items)
#        assert output2 == output1


    def test_filename1(self):
        input = """<ul>
<?py for i in range(0,3): ?>
<li>#{i}</li>
<?py #endfor ?>
</ul>
"""
        filename = 'test_filename1.tenjin'
        try:
            write_file(filename, input)
            template1 = tenjin.Template(filename)
            template2 = tenjin.Template()
            assert template2.convert(input) == template1.script
            assert template2.render() == template1.render()
        finally:
            try:
                os.remove(filename)
            except:
                pass


    def test_import_module1(self):
        import base64
        if python2:
            input = "#{base64.encodestring('tenjin')}"
        elif python3:
            if hasattr(base64, 'encodebytes'):             # python 3.1 or later
                input = "#{base64.encodebytes(b'tenjin')}"
            else:                                          # python 3.0
                input = "#{base64.encodestring(b'tenjin')}"
        template = tenjin.Template()
        template.convert(input)
        def f1():
            template.render()
        with pytest.raises(NameError):
            f1()
        #tenjin.import_module('base64')
        globals()['base64'] = base64
        f1()


    def test_import_module2(self):
        if python2:
            import rfc822
            input = "#{rfc822.formatdate()}"
        elif python3:
            import email.utils
            input = "#{email.utils.formatdate()}"
        template = tenjin.Template()
        template.convert(input)
        def f1():
            template.render()
        with pytest.raises(NameError):
            f1()
        if python2:
            #tenjin.import_module(rfc822)
            globals()['rfc822'] = rfc822
        elif python3:
            globals()['email'] = email
        f1()


    def test_invalid_template_args(self):
        def f():
            input = "<?py #@ARGS 1x ?>"
            template = tenjin.Template()
            template.convert(input)
        with pytest.raises(ValueError):
            f()


#    def test_dummy_if_stmt(self):     ## NEVER!
#        input = r"""
#<html>
#  <body>
#    <ul>
#      <?py if items: ?>
#      <?py   for item in items: ?>
#      <li>#{item}</li>
#      <?py   #endfor ?>
#      <?py #endif ?>
#    </ul>
#  </body>
#</html>
#"""[1:]
#        expected = r"""
#_extend(('''<html>
#  <body>
#    <ul>\n''', ));
#if True: ## dummy
#      if items:
#        for item in items:
#            _extend(('''      <li>''', _to_str(item), '''</li>\n''', ));
#        #endfor
#      #endif
#      _extend(('''    </ul>
#  </body>
#</html>\n''', ));
#"""[1:]
#        t = tenjin.Template()
#        actual = t.convert(input)
#        assert actual == expected

    lvars = "_extend=_buf.extend;_to_str=to_str;_escape=escape; "

    def test_input(self):
        input = r"""<!DOCTYPE>
<ul>
<?py for item in items: ?>
  <li>#{item}</li>
<?py #endfor ?>
</ul>
"""
        script = self.lvars + r"""_extend(('''<!DOCTYPE>
<ul>\n''', ));
for item in items:
    _extend(('''  <li>''', _to_str(item), '''</li>\n''', ));
#endfor
_extend(('''</ul>\n''', ));
"""
        t = tenjin.Template("test.foobar.pyhtml", input=input)
        # input argument is specified then regard it as template content
        assert t.script == script
        # input argument is specified then timestamp is set to False
        assert t.timestamp == False


    def test_trace(self):
        # trace is on then prints template filename as HTML comments
        filename = "test.trace.pyhtml"
        input = "<p>hello #{name}!</p>\n"
        expected = ( "<!-- ***** begin: %s ***** -->\n"
                     "<p>hello world!</p>\n"
                     "<!-- ***** end: %s ***** -->\n" ) % (filename, filename)
        t = tenjin.Template(filename, input=input, trace=True)
        output = t.render({'name':'world'})
        assert output == expected

    def test_option_tostrfunc(self):
        input = "<p>Hello #{name}!</p>"
        # passed tostrfunc option then use it
        globals()['my_str'] = lambda s: s.upper()
        t = tenjin.Template(None, input=input, tostrfunc='my_str')
        output = t.render({'name': 'Haruhi'})
        assert output == "<p>Hello HARUHI!</p>"
        assert t.script == self.lvars.replace('=to_str', '=my_str') + \
                         "_extend(('''<p>Hello ''', _to_str(name), '''!</p>''', ));"
        globals().pop('my_str')
        #
        t = tenjin.Template(None, input=input, tostrfunc='str')
        output = t.render({'name': None})
        assert output == "<p>Hello None!</p>"
        #
        t = tenjin.Template(None, input=input, tostrfunc='str.upper')
        output = t.render({'name': 'sos'})
        assert output == "<p>Hello SOS!</p>"
        # passed False as tostrfunc option then no function is used
        t = tenjin.Template(None, input=input, tostrfunc=False)
        output = t.render({'name': 'Haruhi'})
        assert output == "<p>Hello Haruhi!</p>"
        assert t.script == self.lvars.replace('=to_str', '=False') + \
                         "_extend(('''<p>Hello ''', (name), '''!</p>''', ));"
        #
        def f(): t.render({'name': 123})
        if python2:
            with pytest.raises(TypeError, match=re.escape('sequence item 1: expected string, int found')):
                f()
        elif python3:
            with pytest.raises(TypeError, match=re.escape('sequence item 1: expected str instance, int found')):
                f()
        # passed wrong function name as tostrfunc option then raises error
        t = tenjin.Template(None, input=input, tostrfunc='johnsmith')
        def f(): t.render({'name': 'Haruhi'})
        #with pytest.raises(TypeError, match=re.escape("'NoneType' object is not callable")):
        #    f()
        with pytest.raises(NameError, match=re.escape(_errmsg("name 'johnsmith' is not defined"))):
            f()

    def test_option_escapefunc(self):
        input = "<p>Hello ${name}!</p>"
        # passed escapefunc option then use it
        globals()['my_escape'] = lambda s: s.lower()
        t = tenjin.Template(None, input=input, escapefunc='my_escape')
        output = t.render({'name': 'Haruhi'})
        assert output == "<p>Hello haruhi!</p>"
        assert t.script == self.lvars.replace('=escape', '=my_escape') + \
                         "_extend(('''<p>Hello ''', _escape(_to_str(name)), '''!</p>''', ));"
        globals().pop('my_escape')
        #
        import html as _html_module
        globals()['_html_module'] = _html_module
        t = tenjin.Template(None, input=input, escapefunc='_html_module.escape')
        output = t.render({'name': '&<>"'})
        assert output == "<p>Hello &amp;&lt;&gt;&quot;!</p>"
        # passed False as escapefunc option then no function is used
        t = tenjin.Template(None, input=input, escapefunc=False)
        output = t.render({'name': 'Haru&Kyon'})
        assert output == "<p>Hello Haru&Kyon!</p>"
        assert t.script == self.lvars.replace('=escape', '=False') + \
                         "_extend(('''<p>Hello ''', _to_str(name), '''!</p>''', ));"
        # passed wrong function name as tostrfunc option then raises error
        t = tenjin.Template(None, input=input, escapefunc='kyonsmith')
        def f(): t.render({'name': 'Haruhi'})
        #with pytest.raises(TypeError, match=re.escape("'NoneType' object is not callable")):
        #    f()
        with pytest.raises(NameError, match=re.escape(_errmsg("name 'kyonsmith' is not defined"))):
            f()

    def test_localvars_assignments_without_args_declaration(self):
        def _convert(input):
            return tenjin.Template(input=input).script
        # add local vars assignments before text only once.
        input = r"""
<p>
  <?py x = 10 ?>
</p>
"""[1:]
        expected = r"""
_extend=_buf.extend;_to_str=to_str;_escape=escape; _extend(('''<p>\n''', ));
x = 10
_extend(('''</p>\n''', ));
"""[1:]
        assert _convert(input) == expected
        # skips comments at the first lines.
        input = r"""
<?py # coding: utf-8 ?>
<?py      ### comment ?>
<p>
  <?py x = 10 ?>
</p>
"""[1:]
        expected = r"""
# coding: utf-8
### comment
_extend=_buf.extend;_to_str=to_str;_escape=escape; _extend(('''<p>\n''', ));
x = 10
_extend(('''</p>\n''', ));
"""[1:]
        assert _convert(input) == expected
        # adds local vars assignments before statements.
        input = r"""
<?py for item in items: ?>
  <?py x = 10 ?>
<?py #endfor ?>
</p>
"""[1:]
        expected = r"""
_extend=_buf.extend;_to_str=to_str;_escape=escape; 
for item in items:
    x = 10
#endfor
_extend(('''</p>\n''', ));
"""[1:]
        assert _convert(input) == expected

    def test_localvars_assignments_with_args_declaration(self):
        def _convert(input):
            return tenjin.Template(input=input).script
        # args declaration exists before text then local vars assignments apprears at the same line with text.
        input = r"""
<?py # coding: utf-8 ?>
<?py #@ARGS items ?>
<p>
  <?py x = 10 ?>
</p>
"""[1:]
        expected = r"""
# coding: utf-8
items = _context.get('items'); 
_extend=_buf.extend;_to_str=to_str;_escape=escape; _extend(('''<p>\n''', ));
x = 10
_extend(('''</p>\n''', ));
"""[1:]
        assert _convert(input) == expected
        # args declaration exists before statement then local vars assignments apprears at the same line with args declaration.
        input = r"""
<?py # coding: utf-8 ?>
<?py #@ARGS items ?>
  <?py x = 10 ?>
"""[1:]
        expected = r"""
# coding: utf-8
_extend=_buf.extend;_to_str=to_str;_escape=escape; items = _context.get('items'); 
x = 10
"""[1:]
        assert _convert(input) == expected
        # 'from __future__' statement exists then skip it.
        input = r"""
<?py from __future__ import with_statement ?>
<?py # coding: utf-8 ?>
<?py #@ARGS item ?>
item=${item}
"""[1:]
        expected = r"""
from __future__ import with_statement
# coding: utf-8
item = _context.get('item'); 
_extend=_buf.extend;_to_str=to_str;_escape=escape; _extend(('''item=''', _escape(_to_str(item)), '''\n''', ));
"""[1:]
        assert _convert(input) == expected
        input = r"""
<?py from __future__ import with_statement ?>
<?py for item in items: ?>
  <p>${item}</p>
<?py #endfor ?>
"""[1:]
        expected = r"""
from __future__ import with_statement
_extend=_buf.extend;_to_str=to_str;_escape=escape; 
for item in items:
    _extend(('''  <p>''', _escape(_to_str(item)), '''</p>\n''', ));
#endfor
"""[1:]
        assert _convert(input) == expected

    def test_new_notation(self):
        input = r"""
a=${a}
b=#{b}
c={=c=}
d={==d==}
"""
        expected = r"""_extend=_buf.extend;_to_str=to_str;_escape=escape; _extend(('''
a=''', _escape(_to_str(a)), '''
b=''', _to_str(b), '''
c=''', _escape(_to_str(c)), '''
d=''', _to_str(d), '''\n''', ));
"""
        t = tenjin.Template()
        assert t.convert(input) == expected


    def test_add_expr(self):
        input = r"""
not escape: #{var}
escape: ${var}
"""
        # nothing is specified then both _to_str() and _escape() are used.
        t = tenjin.Template()
        script = t.convert(input)
        expected = r"""_extend=_buf.extend;_to_str=to_str;_escape=escape; _extend(('''
not escape: ''', _to_str(var), '''
escape: ''', _escape(_to_str(var)), '''\n''', ));
"""
        assert script == expected
        # when tostrfunc is False then skips _to_str().
        expected = r"""_extend=_buf.extend;_to_str=False;_escape=escape; _extend(('''
not escape: ''', (var), '''
escape: ''', _escape(var), '''\n''', ));
"""
        t = tenjin.Template(tostrfunc=False)
        assert t.convert(input) == expected
        # escapefunc is False then skips _escape().
        expected = r"""_extend=_buf.extend;_to_str=to_str;_escape=False; _extend(('''
not escape: ''', _to_str(var), '''
escape: ''', _to_str(var), '''\n''', ));
"""
        t = tenjin.Template(escapefunc=False)
        assert t.convert(input) == expected
        # both tostr and escapefunc are False then skips _to_str() and _escape().
        expected = r"""_extend=_buf.extend;_to_str=False;_escape=False; _extend(('''
not escape: ''', (var), '''
escape: ''', (var), '''\n''', ));
"""
        t = tenjin.Template(tostrfunc=False, escapefunc=False)
        assert t.convert(input) == expected
        # get_expr_and_flags() returns flag_tostr=False then ignores _escape().
        expected = r"""_extend=_buf.extend;_to_str=to_str;_escape=escape; _extend(('''
not escape: ''', (var), '''
escape: ''', _escape(var), '''\n''', ));
"""
        t = tenjin.Template()
        _orig_get_expr = t.get_expr_and_flags
        def fn(*args):
            expr, (flag_escape, flag_tostr) = _orig_get_expr(*args)
            return expr, (flag_escape, False)
        t.get_expr_and_flags = fn
        assert t.convert(input) == expected
        # get_expr_and_flags() returns flag_escape=False then ignores _escape().
        expected = r"""_extend=_buf.extend;_to_str=to_str;_escape=escape; _extend(('''
not escape: ''', _to_str(var), '''
escape: ''', _to_str(var), '''\n''', ));
"""
        t = tenjin.Template()
        _orig_get_expr = t.get_expr_and_flags
        def fn(*args):
            expr, (flag_escape, flag_tostr) = _orig_get_expr(*args)
            return expr, (False, flag_tostr)
        t.get_expr_and_flags = fn
        assert t.convert(input) == expected
        # get_expr_and_flags() returns both flags False then ignores both _to_str() and _escape().
        expected = r"""_extend=_buf.extend;_to_str=to_str;_escape=escape; _extend(('''
not escape: ''', (var), '''
escape: ''', (var), '''\n''', ));
"""
        t = tenjin.Template()
        _orig_get_expr = t.get_expr_and_flags
        def fn(*args):
            expr, (flag_escape, flag_tostr) = _orig_get_expr(*args)
            return expr, (False, False)
        t.get_expr_and_flags = fn
        assert t.convert(input) == expected

    def test_new_cycle(self):
        cycle = tenjin.helpers.new_cycle('odd', 'even')
        assert cycle() == 'odd'
        assert cycle() == 'even'
        assert cycle() == 'odd'
        assert cycle() == 'even'
        #
        cycle = tenjin.helpers.new_cycle('A', 'B', 'C')
        assert cycle() == 'A'
        assert cycle() == 'B'
        assert cycle() == 'C'
        assert cycle() == 'A'
        assert cycle() == 'B'
        assert cycle() == 'C'
        #
        #assert isinstance(cycle(), EscapedStr)
        #assert isinstance(cycle(), EscapedStr)

    def test_curly_braces_in_expressions(self):
        # '{}' is available in '${}' or '#{}', such as '${foo({'x':1})}'
        input = """
<p>${f({'a':1})+g({'b':2})}</p>
<p>#{f({'c':3})+g({'d':4})}</p>
"""
        expected = r"""_extend=_buf.extend;_to_str=to_str;_escape=escape; _extend(('''
<p>''', _escape(_to_str(f({'a':1})+g({'b':2}))), '''</p>
<p>''', _to_str(f({'c':3})+g({'d':4})), '''</p>\n''', ));
"""
        t = tenjin.Template()
        script = t.convert(input)
        assert script == expected



###
### $Release: 1.0.0 $
### Copyright (c) 2024-present Hyun-Gyu Kim (babyworm@gmail.com). MIT License.
### Original: copyright(c) 2007-2012 kuwata-lab.com all rights reserved.
###

import pytest
import re as _re
import sys, os, re, time, marshal, shutil
from glob import glob
try:    import cPickle as pickle
except: import pickle

from testcase_helper import *
import tenjin
#from tenjin.helpers import escape, to_str
from tenjin.helpers import *

lvars = "_extend=_buf.extend;_to_str=to_str;_escape=escape; "

JYTHON = hasattr(sys, 'JYTHON_JAR')


def _convert_data(data, lang='python'):
    if isinstance(data, dict):
        for k in list(data.keys()):
            v = data[k]
            if k[-1] == '*':
                assert isinstance(v, dict)
                data[k[:-1]] = v.get(lang)
            if isinstance(v, dict) and lang in v:
                data[k] = v[lang]
            else:
                _convert_data(v, lang)
    elif isinstance(data, list):
        for k, v in enumerate(data):
            if isinstance(v, dict) and lang in v:
                data[k] = v[lang]
            else:
                _convert_data(v, lang)


def _remove_files(basenames=[]):
    for basename in basenames:
        for filename in glob("%s*" % basename):
            os.unlink(filename)


def _with_dummy_files(func):
    """decorator for test functions"""
    def deco(self):
        isdir = os.path.isdir
        mkdir = os.mkdir
        try:
            if not isdir('_views'): mkdir('_views')
            if not isdir('_views/blog'): mkdir('_views/blog')
            pairs = [ ('_views/blog/index.pyhtml', 'xxx'),
                      ('_views/index.pyhtml', '<<#{{_DUMMY_VALUE}}>>'),
                      ('_views/layout.pyhtml', '<div>#{_content}</div>'), ]
            for fname, cont in pairs:
                f = open(fname, 'w')
                f.write(cont)
                f.close()
            func(self)
        finally:
            shutil.rmtree('_views')
    for x in ('func_name', '__name__', '__doc__'):
        if hasattr(func, x):
            setattr(deco, x, getattr(func, x))
    return deco


class DebugLogger(object):

    def __init__(self):
        self.messages = []

    def __str__(self):
        return "\n".join(self.messages)

    def f(name):
        def func(self, format, *args):
            msg = args and (format % args) or format
            self.messages.append("[%s] %s" % (name.upper(), msg))
        func.func_name = func.__name__ = name
        return func

    fatal, error, warn, info, debug, trace = \
        f('fatal'), f('error'), f('warn'), f('info'), f('debug'), f('trace')

    del f


class TestEngine:

    #code = TestCaseHelper.generate_testcode(__file__)
    #exec(code)
    datalist = TestCaseHelper.load_testdata(__file__)
    _testdata = dict([ (d['name'], d) for d in datalist ])
    _convert_data(_testdata, lang='python')
    data = _testdata['basic']
    for d in data['templates']:
        d['filename'] = d['filename'].replace('.xxhtml', '.pyhtml')
    templates = dict([(hash['filename'], hash['content']) for hash in data['templates']])
    expected  = dict([(hash['name'], hash['content']) for hash in data['expected']])
    contexts  = data['contexts']


    def setup_method(self):
        tenjin.Engine.cache.clear()
    #    testdata = TestEngine._testdata['basic']
    #    for hash in testdata['templates']:
    #        write_file(hash['filename'], hash['content'])


    #def teardown_method(self):
    #    for hash in TestEngine._testdata['basic']['templates']:
    #        filename = hash['filename']
    #        for fname in [filename, filename+'.cache', filename+'.marshal']:
    #            if os.path.exists(fname):
    #                os.unlink(fname)


    def _testname(self):
        try:
            return self._TestCase__testMethodName
        except AttributeError:
            return self._testMethodName

    def _test_basic(self):
        try:
            testdata = TestEngine._testdata['basic']
            for hash in testdata['templates']:
                write_file(hash['filename'], hash['content'])
            #
            testname = self._testname()
            lst = testname[len('test_basic'):].split('_')
            action   = lst[1]                # 'list', 'show', 'create', or 'edit'
            shortp   = lst[2] == 'short'     # 'short' or 'long'
            layoutp  = lst[3] != 'nolayout'  # 'nolayout' or 'withlayout'
            layout   = layoutp and 'user_layout.pyhtml' or None
            engine   = tenjin.Engine(prefix='user_', postfix='.pyhtml', layout=layout)
            context  = TestEngine.contexts[action].copy()
            key      = 'user_' + action + (layout and '_withlayout' or '_nolayout')
            expected = TestEngine.expected[key]
            filename = 'user_%s.pyhtml' % action
            tplname  = shortp and ':'+action or filename
            if layout:
                output = engine.render(tplname, context)
            else:
                output = engine.render(tplname, context, layout=False)
            assert output == expected
        finally:
            filenames = [ hash['filename'] for hash in TestEngine._testdata['basic']['templates'] ]
            _remove_files(filenames)
            #for hash in TestEngine._testdata['basic']['templates']:
            #    filename = hash['filename']
            #    for fname in [filename, filename+'.cache', filename+'.marshal']:
            #        if os.path.exists(fname):
            #            os.unlink(fname)


    ## long, nolayout

    def test_basic_list_long_nolayout(self):
        self._test_basic()

    def test_basic_show_long_nolayout(self):
        self._test_basic()

    def test_basic_create_long_nolayout(self):
        self._test_basic()

    def test_basic_edit_long_nolayout(self):
        self._test_basic()


    ## short, nolayout

    def test_basic_list_short_nolayout(self):
        self._test_basic()

    def test_basic_show_short_nolayout(self):
        self._test_basic()

    def test_basic_create_short_nolayout(self):
        self._test_basic()

    def test_basic_edit_short_nolayout(self):
        self._test_basic()


    ## long, withlayout

    def test_basic_list_long_withlayout(self):
        self._test_basic()

    def test_basic_show_long_withlayout(self):
        self._test_basic()

    def test_basic_create_long_withlayout(self):
        self._test_basic()

    def test_basic_edit_long_withlayout(self):
        self._test_basic()


    ## short, withlayout

    def test_basic_list_short_withlayout(self):
        self._test_basic()

    def test_basic_show_short_withlayout(self):
        self._test_basic()

    def test_basic_create_short_withlayout(self):
        self._test_basic()

    def test_basic_edit_short_withlayout(self):
        self._test_basic()


    ## ----------------------------------------


    def test_capture_and_echo(self):
        hash = TestEngine._testdata['test_capture_and_echo']
        layout = hash['layout']
        content = hash['content']
        expected = hash['expected']
        layout_filename = 'user_layout.pyhtml'
        content_filename = 'user_content.pyhtml'
        context = { 'items': ['AAA', 'BBB', 'CCC'] }
        try:
            write_file(layout_filename, layout)
            write_file(content_filename, content)
            engine = tenjin.Engine(prefix='user_', postfix='.pyhtml', layout=':layout')
            output = engine.render(':content', context)
            assert output == expected
        finally:
            _remove_files([layout_filename, content_filename])


    def test_capture_as(self):
        filename  = '.test.capture_as.pyhtml'
        content1  = ("<?py from __future__ import with_statement ?>\n"
                     "<div>\n"
                     "<?py with capture_as('sidemenu'): ?>\n"
                     "  <ul>\n"
                     "  <?py for item in items: ?>\n"
                     "    <li>${item}</li>\n"
                     "  <?py #endfor ?>\n"
                     "  </ul>\n"
                     "<?py #endwith ?>\n"
                     "</div>\n")
        content2  = ("<div>\n"
                     "<?py for _ in capture_as('sidemenu'): ?>\n"
                     "  <ul>\n"
                     "  <?py for item in items: ?>\n"
                     "    <li>${item}</li>\n"
                     "  <?py #endfor ?>\n"
                     "  </ul>\n"
                     "<?py #endfor ?>\n"
                     "</div>\n")
        expected1 = ("<div>\n"
                     "</div>\n")
        expected2 = ("  <ul>\n"
                     "    <li>A</li>\n"
                     "    <li>B</li>\n"
                     "    <li>C</li>\n"
                     "  </ul>\n")
        contents = (content1, content2)
        if sys.version < '2.5':
            contents = (content2, )
        for content in contents:
            # Replace dummy_file usage with manual file creation/cleanup
            f = open(filename, 'w'); f.write(content); f.close()
            try:
                context = {'items': ['A', 'B', 'C']}
                engine = tenjin.Engine()
                output = engine.render(filename, context)
                assert output == expected1
                assert context['sidemenu'] == expected2
            finally:
                if os.path.exists(filename): os.unlink(filename)

    def test_captured_as(self):
        hash = TestEngine._testdata['test_captured_as']
        files = ( ('content.pyhtml',      hash['content']),
                  ('customlayout.pyhtml', hash['customlayout']),
                  ('baselayout.pyhtml',   hash['baselayout']),
                  )
        context = hash['context']
        expected = hash['expected']
        try:
            for filename, content in files:
                write_file(filename, content)
            engine = tenjin.Engine(postfix='.pyhtml')
            output = engine.render(':content', context)
            assert output == expected
        finally:
            _remove_files([ t[0] for t in files ])


    def test_local_layout(self):
        hash = TestEngine._testdata['test_local_layout']
        context = hash['context']
        names = ['layout_html', 'layout_xhtml', 'content_html']
        def fname(base):
            return 'local_%s.pyhtml' % base
        try:
            interval = tenjin.Engine.timestamp_interval
            tenjin.Engine.timestamp_interval = 0
            for name in names:
                write_file(fname(name), hash[name])
            engine = tenjin.Engine(prefix='local_', postfix='.pyhtml', layout=':layout_html')
            ##
            def _test(expected, statement):
                content_html = hash['content_html'] + statement
                write_file(fname('content_html'), content_html)
                actual = engine.render(':content_html', context)
                assert actual == expected
            ##
            _test(hash['expected_html'], '')
            time.sleep(1)
            _test(hash['expected_xhtml'], "<?py _context['_layout'] = ':layout_xhtml' ?>\n")
            time.sleep(1)
            _test(hash['expected_nolayout'], "<?py _context['_layout'] = False ?>\n")
            ##
        finally:
            tenjin.Engine.timestamp_interval = interval
            #for name in names:
            #    for suffix in ['', '.cache', '.marshal']:
            #        filename = fname(name) + suffix
            #        if os.path.isfile(filename):
            #            os.unlink(filename)
            _remove_files([ fname(name) for name in names ])


    def test_cachefile(self):
        data = TestEngine._testdata['test_cachefile']
        filenames = { 'layout': 'layout.pyhtml',
                      'page': 'account_create.pyhtml',
                      'form': 'account_form.pyhtml',
                    }
        expected = data['expected']
        context = { 'params': { } }
        cache_filenames = ['account_create.pyhtml.cache', 'account_form.pyhtml.cache']
        try:
            for key, filename in filenames.items():
                write_file(filename, data[key])
            props = { 'prefix':'account_', 'postfix':'.pyhtml', 'layout':'layout.pyhtml' }
            ## no caching
            props['cache'] = False
            engine = tenjin.Engine(**props)
            output = engine.render(':create', context)
            assert output == expected
            for fname in cache_filenames: assert not os.path.exists(fname)
            ## marshal caching
            if not JYTHON:
                props['cache'] = tenjin.MarshalCacheStorage()
                engine = tenjin.Engine(**props)
                output = engine.render(':create', context)
                assert output == expected
                if   python2:  nullbyte = '\0'
                elif python3:  nullbyte = '\0'.encode('ascii')
                for fname in cache_filenames:
                    assert os.path.exists(fname)               # file created?
                    s = read_file(fname, 'rb')        # read binary file
                    assert s.find(nullbyte) >= 0        # binary file?
                    f = open(fname, 'rb')
                    fn = lambda: marshal.load(f)
                    try:
                        fn()           # marshal?
                    finally:
                        f.close()
                engine = tenjin.Engine(**props)
                output = engine.render(':create', context)
                assert output == expected               # reloadable?
            #
            for fname in glob('*.pyhtml.cache'): os.unlink(fname)
            for fname in cache_filenames:
                assert not os.path.exists(fname)
            ## pickle caching
            props['cache'] = tenjin.PickleCacheStorage()
            engine = tenjin.Engine(**props)
            output = engine.render(':create', context)
            assert output == expected
            if   python2:  nullbyte = '\0'
            elif python3:  nullbyte = '\0'.encode('ascii')
            for fname in cache_filenames:
                assert os.path.exists(fname)                         # file created?
                s = read_file(fname, 'rb')                       # read text file
                if python2:
                    assert s.find(nullbyte) < 0        # text file? (pickle protocol ver 2)
                elif python3:
                    assert s.find(nullbyte) >= 0       # binary file? (pickle protocol ver 3)
                f = open(fname, 'rb')
                fn = lambda: pickle.load(f)
                try:
                    fn()
                finally:
                    f.close()
                f = open(fname, 'rb')
                pickle.load(f)
                f.close()
            engine = tenjin.Engine(**props)
            output = engine.render(':create', context)
            assert output == expected               # reloadable?
            #
            for fname in glob('*.cache'): os.unlink(fname)
            for fname in cache_filenames:
                assert not os.path.exists(fname)
            ## text caching
            props['cache'] = tenjin.TextCacheStorage()
            engine = tenjin.Engine(**props)
            output = engine.render(':create', context)
            assert output == expected
            if   python2:  nullchar = '\0'
            elif python3:  nullchar = '\0'
            for fname in cache_filenames:
                assert os.path.exists(fname)                  # file created?
                s = read_file(fname, 'r')            # read text file
                assert s.find(nullchar) < 0            # text file?
                if JYTHON:
                    continue
                #fn = lambda: marshal.loads(s)
                f = open(fname, 'rb')
                fn = lambda: marshal.load(f)
                try:
                  if python3:
                    with pytest.raises((ValueError, EOFError)) as exc_info: fn()        # non-marshal?
                    ex = exc_info.value
                    if sys.version_info[1] == 0:     # python 3.0
                        assert str(ex) == "bad marshal data"
                    else:                            # python 3.1 or later
                        if isinstance(ex, ValueError):
                            assert str(ex) == "bad marshal data (unknown type code)"
                        # EOFError for Python 3.12+ is also acceptable
                  elif python2 and sys.version_info[1] >= 5:
                    with pytest.raises(EOFError, match=_re.escape("EOF read where object expected")): fn()  # non-marshal?
                finally:
                    f.close()
            engine = tenjin.Engine(**props)
            output = engine.render(':create', context)
            assert output == expected                  # reloadable?
        finally:
            _remove_files(filenames.values())

    def test_cachefile_timestamp(self):
        """engine should clear cache not only template is newer but also template is older than cache."""
        data = TestEngine._testdata['test_cachefile']
        filenames = { 'layout': 'layout.pyhtml',
                      'page': 'account_create.pyhtml',
                      'form': 'account_form.pyhtml',
                    }
        expected = data['expected']
        context = { 'params': { } }
        cache_filenames = ['account_create.pyhtml.cache', 'account_form.pyhtml.cache']
        try:
            interval = tenjin.Engine.timestamp_interval
            tenjin.Engine.timestamp_interval = 0
            for key, filename in filenames.items():
                write_file(filename, data[key])
            props = { 'prefix': 'account_', 'postfix':'.pyhtml', 'layout':'layout.pyhtml', 'cache':True }
            ## create cache files and check them
            time.sleep(1)
            curr_time = time.time()
            engine = tenjin.Engine(**props)
            output = engine.render(':create', context)
            for fname in filenames.values():
                assert os.path.exists(fname)                         # file created?
                assert engine.get_template(fname).timestamp < curr_time
                assert engine.get_template(fname).timestamp == os.path.getmtime(fname)
            ## save current cached object
            cached = {}
            for fname in filenames.values():
                cached[fname] = engine.get_template(fname)
            ## confirm that get_template() returns the same object
            for fname in filenames.values():
                assert cached[fname] is engine.get_template(fname)
            ## change timestamp of templates to be old
            for fname in filenames.values():
                atime = mtime = os.path.getmtime(fname) - 10
                os.utime(fname, (atime, mtime))
            ## check whether new caches are created
            for fname in filenames.values():
                t = engine.get_template(fname)
                assert cached[fname] is not t
                assert t.timestamp == os.path.getmtime(fname)
        finally:
            tenjin.Engine.timestamp_interval = interval
            _remove_files(filenames.values())


    def test_cachefile_unset(self):
        try:
            input = "<?py x = 10 ?>"
            template_name = "cachefile_delete.pyhtml"
            f = open(template_name, 'w'); f.write(input); f.close()
            storage = tenjin.MarshalCacheStorage()
            engine = tenjin.Engine(cache=storage)
            engine.render(template_name)
            fullpath = os.path.abspath(template_name)
            cachepath = engine.cachename(fullpath)
            assert cachepath in storage.items
            assert os.path.exists(template_name + '.cache')
            storage.unset(cachepath)
            assert cachepath not in storage.items
            assert not os.path.exists(template_name + '.cache')
            def f():
                storage.unset(fullpath)
            f()
        finally:
            _remove_files([template_name, template_name+'.cache'])


    def test_change_layout(self):
        data = TestEngine._testdata['test_change_layout']
        ## setup
        tenjin.Engine.cache.clear()
        basenames = ['baselayout', 'customlayout', 'content']
        for basename in basenames:
            write_file('%s.pyhtml' % basename, data[basename])
        ## body
        try:
            engine = tenjin.Engine(layout='baselayout.pyhtml')
            output = engine.render('content.pyhtml')
            expected = data['expected']
            assert output == expected
        ## teardown
        finally:
            _remove_files(basenames)


    def test_context_scope(self):
        data = TestEngine._testdata['test_context_scope']
        base = data['base']
        part = data['part']
        expected = data['expected']
        for basename in ('base', 'part'):
            write_file('%s.pyhtml' % basename, data[basename])
        #
        try:
            engine = tenjin.Engine()
            context = {}
            output = engine.render('base.pyhtml', context)
            expected = data['expected']
            assert output == expected
        finally:
            _remove_files(['base', 'part'])


    def test_template_args(self):
        data = TestEngine._testdata['test_template_args']
        content = data['content']
        expected = data['expected']
        context = data['context']
        for basename in ('content', ):
            write_file('%s.pyhtml' % basename, data[basename])
        #
        try:
            def f1():
                engine = tenjin.Engine(cache=True)
                assert engine.get_template('content.pyhtml').args != None
                output = engine.render('content.pyhtml', context)
            with pytest.raises(NameError) as exc_info: f1()
            ex = exc_info.value
            #import sys; sys.stderr.write("*** debug: ex=%s\n" % (repr(ex)))
            #engine = tenjin.Engine(cache=True)
            #assert engine.get_template('content.pyhtml').args != None
            #output = engine.render('content.pyhtml', context)
            #assert output == expected
            with pytest.raises(NameError) as exc_info: f1()
            ex = exc_info.value
        finally:
            _remove_files(['content'])


    def test__set_cache_storage(self):
        if "default then Engine.cache is TextCacheStorage instance":
            assert isinstance(tenjin.Engine.cache, tenjin.TextCacheStorage)
        if "cache=True specified then use default cache object":
            engine = tenjin.Engine(cache=True)
            assert ('cache' in engine.__dict__) == False
            assert engine.cache is tenjin.Engine.cache
            assert engine.cache is tenjin.Engine(cache=True).cache
        if "cache=True and default cache is not set then create MarshalCacheStorage object for each engine":
            bkup = tenjin.Engine.cache
            try:
                tenjin.Engine.cache = None
                engine = tenjin.Engine(cache=True)
                assert 'cache' in engine.__dict__
                assert isinstance(engine.cache, tenjin.MarshalCacheStorage)
                assert engine.cache is not tenjin.Engine(cache=True).cache
            finally:
                tenjin.Engine.cache = bkup
        #if "cache=None specified then set MemoryCacheObject instance as cache object":
        #    engine = tenjin.Engine(cache=None)
        #    assert ('cache' in engine.__dict__) == True
        #    assert isinstance(engine.cache, tenjin.MemoryCacheStorage)
        if "cache=None then do nothing":
            engine = tenjin.Engine(cache=None)
            assert 'cache' not in engine.__dict__
        if "cache=False specified then don't use cache object":
            engine = tenjin.Engine(cache=False)
            assert 'cache' in engine.__dict__
            assert engine.cache == None
        if "CacheStorage instance is specified then use it as cache object":
            cache_storage = tenjin.MarshalCacheStorage()
            engine = tenjin.Engine(cache=cache_storage)
            assert 'cache' in engine.__dict__
            assert engine.cache is cache_storage
        if "invalid object is specified as cache object then raise ValueError":
            def f():
                tenjin.Engine(cache=123)
            with pytest.raises(ValueError, match=_re.escape('123: invalid cache object.')): f()


    def test_cached_contents(self):
        data = TestEngine._testdata['test_cached_contents']
        def _test(filename, cachename, cachemode, input, expected_script, expected_args):
            if input:
                write_file(filename, input)
            engine = tenjin.Engine(cache=cachemode)
            t = engine.get_template(filename)
            assert t.args == expected_args
            assert t.script == expected_script
            #import marshal
            #f = open(filename + '.cache', 'rb')
            #try:
            #    dct = marshal.load(f)
            #    assert dct['args'] == expected_args
            #    assert dct['script'] == expected_script
            #finally:
            #    f.close()
        ##
        try:
            ## args=[x,y,z], cache=1
            filename = 'input.pyhtml'
            for f in glob(filename+'*'): os.path.exists(f) and os.remove(f)
            script = data['script1']
            args  = data['args1']
            input = data['input1']
            cachename = filename+'.cache'
            assert not os.path.exists(cachename)
            _test(filename, cachename, True, input, script, args)
            assert os.path.exists(cachename)
            _test(filename, cachename, True, None, script, args)
            ## args=[], cache=1
            cachename = filename+'.cache'
            input = data['input2']  # re.sub(r'<\?py #@ARGS.*?\?>\n', '<?py #@ARGS ?>\n', input)
            script = data['script2']  # re.sub(r'#@ARGS.*?\n', '#@ARGS \n', cache)
            args  = data['args2']   # []
            time.sleep(1)
            #assert os.path.exists(cachename)
            _test(filename, cachename, True, input, script, args)
            #assert os.path.exists(cachename)
            _test(filename, cachename, True, None, script, args)
        finally:
            _remove_files(['input.pyhtml'])


    def _test_template_path(self, keys):
        data = TestEngine._testdata['test_template_path']
        basedir = 'test_templates'
        try:
            os.mkdir(basedir)
            os.mkdir(basedir + '/common')
            os.mkdir(basedir + '/user')
            d = { 'layout':keys[0], 'body':keys[1], 'footer':keys[2], }
            for key in ('layout', 'body', 'footer'):
                filename = '%s/common/%s.pyhtml' % (basedir, key)
                write_file(filename, data['common_'+key])
                if d[key] == 'user':
                    filename = '%s/user/%s.pyhtml' % (basedir, key)
                    write_file(filename, data['user_'+key])
            #
            path = [basedir+'/user', basedir+'/common']
            engine = tenjin.Engine(postfix='.pyhtml', path=path, layout=':layout')
            context = {'items':('AAA', 'BBB', 'CCC')}
            output = engine.render(':body', context)
            #
            expected = data['expected_' + '_'.join(keys)]
            assert output == expected
        finally:
            #os.removedirs(basedir)
            #pass
            for filename in glob('%s/*/*' % basedir):
                os.unlink(filename)
            for filename in glob('%s/*' % basedir):
                os.rmdir(filename)
            os.rmdir(basedir)

    def test_template_path_common_common_common(self):
        self._test_template_path(('common', 'common', 'common'))
    def test_template_path_user_common_common(self):
        self._test_template_path(('user',   'common', 'common'))
    def test_template_path_common_user_common(self):
        self._test_template_path(('common', 'user',   'common'))
    def test_template_path_user_user_common(self):
        self._test_template_path(('user',   'user',   'common'))
    def test_template_path_common_common_user(self):
        self._test_template_path(('common', 'common', 'user'))
    def test_template_path_user_common_user(self):
        self._test_template_path(('user',   'common', 'user'))
    def test_template_path_common_user_user(self):
        self._test_template_path(('common', 'user',   'user'))
    def test_template_path_user_user_user(self):
        self._test_template_path(('user',   'user',   'user'))


    def test_preprocessor(self):
        data = TestEngine._testdata['test_preprocessor']
        try:
            basenames = ('form', 'create', 'update', 'layout', )
            filenames = []
            for name in basenames:
                filename = 'prep_%s.pyhtml' % name
                filenames.append(filename)
                write_file(filename, data[name])
            engine = tenjin.Engine(prefix='prep_', postfix='.pyhtml', layout=':layout', preprocess=True)
            #
            context = {
                'title': 'Create',
                'action': 'create',
                'params': { 'state': 'NY' },
            }
            actual = engine.render(':create', context)  # 1st
            assert actual == data['expected1']
            context['params'] = {'state': 'xx'}
            actual = engine.render(':create', context)  # 2nd
            #assert actual == data['expected1']
            expected = data['expected1'].replace(r' checked="checked"', '')
            assert actual == expected
            #
            context = {
                'title': 'Update',
                'action': 'update',
                'params': { 'state': 'NY' },
            }
            actual = engine.render(':update', context)  # 1st
            assert actual == data['expected2']
            context['params'] = { 'state': 'xx' }
            actual = engine.render(':update', context)  # 2nd
            assert actual == data['expected2'] # not changed!
            #
            assert engine.get_template(':form').script == data['cache1']
            assert engine.get_template(':create').script == data['cache2']
            assert engine.get_template(':layout').script == data['cache3']
            assert engine.get_template(':update').script == data['cache4']
            #
        finally:
            _remove_files(filenames)

    @_with_dummy_files
    def test_pp(self):
        # 'pp' paramater should be a list of preprocessor objects.
        pp1 = tenjin.TemplatePreprocessor()
        pp2 = tenjin.TrimPreprocessor()
        pp3 = tenjin.JavaScriptPreprocessor(type="text/javascript")
        e = tenjin.Engine(pp=[pp1, pp2, pp3])
        input = r"""
<body>
  <div>
    <!-- #JS: render_items(items) -->
    <ul>
    <?js for (var i = 0; i < items.length; i++) { ?>
      <li>${i}</li>
    <?js } ?>
    </ul>
    <!-- #/JS -->
  </div>
  <script>#{{tenjin.JS_FUNC}}</script>
</body>
"""[1:]
        expected = r"""
<body>
<div>
<script type="text/javascript">function render_items(items){var _buf='';
_buf+='<ul>\n';
 for (var i = 0; i < items.length; i++) {
_buf+='<li>'+_E(i)+'</li>\n';
 }
_buf+='</ul>\n';
return _buf;};</script>
</div>
<script>function _S(x){return x==null?'':x;}
function _E(x){return x==null?'':typeof(x)!=='string'?x:x.replace(/[&<>"']/g,_EF);}
var _ET={'&':"&amp;",'<':"&lt;",'>':"&gt;",'"':"&quot;","'":"&#039;"};
function _EF(c){return _ET[c];};</script>
</body>
"""[1:]
        fname = 'tmp_123.pyhtml'
        f = open(fname, 'w'); f.write(input); f.close()
        try:
            t = e.get_template(fname)
            context = {}
            output = e.render(fname, context)
            assert output == expected
        finally:
            for x in glob(fname + '*'): os.unlink(x)

    def test_init_creates_templatepreprocessor_object_when_preprocess_option_is_on(self):
        # #__init__(): creates TemplatePreprocessor object when 'preprocess' option is on.
        e = tenjin.Engine(preprocess=True)
        assert isinstance(e.pp, list); assert len(e.pp) == 1
        assert isinstance(e.pp[0], tenjin.TemplatePreprocessor)

    def test_init_creates_templatepreprocessor_object_with_preprocessorclass_class(self):
        # #__init__(): creates TemplatePreprocessor object with 'preprocessorclass' class.
        e = tenjin.Engine(preprocess=True)
        assert e.pp[0].factory == tenjin.Preprocessor
        e = tenjin.Engine(preprocess=True, preprocessorclass=tenjin.SafePreprocessor)
        assert e.pp[0].factory == tenjin.SafePreprocessor

    def test_include_with_kwargs(self):
        data = TestEngine._testdata['test_include_with_kwargs']
        write_file('index.pyhtml', data['index_html'])
        write_file('sub.pyhtml', data['sub_html'])
        expected = data['expected']
        #
        try:
            engine = tenjin.Engine()
            context = {}
            output = engine.render('index.pyhtml', context)
            assert output == expected
        finally:
            _remove_files(['index', 'sub'])


    def test_add_template(self):
        if "template is added then it can be got by get_template()":
            input = """val=#{val}"""
            template = tenjin.Template('foo.pyhtml', input=input)
            engine = tenjin.Engine(postfix='.pyhtml')
            engine.add_template(template)
            assert engine.get_template('foo.pyhtml') == template
            assert engine.get_template(':foo') == template
        if "template is added then it should not create cache file":
            assert engine.render(':foo', {'val': 'ABC'}) == 'val=ABC'
            assert not os.path.exists('foo.pyhtml')


    #def test_prefer_fullpath(self):
    #    fname = 'test_prefer_fullpath.pyhtml'
    #    input = ( "<?py for x in items: ?>\n"
    #              "  <p>${x.length}</p>\n"
    #              "<?py #end ?>\n" )
    #    context = {'items': ['A']}
    #    engine = None
    #    #
    #    def _test(fname, input, expected):
    #        write_file(fname, input)
    #        engine = tenjin.Engine()
    #        engine.cache.clear()
    #        t = engine.get_template(fname)
    #        assert t.filename == expected
    #        ex = error_file = None
    #        try:
    #            engine.render(fname, context)
    #        except:
    #            ex = sys.exc_info()[1]
    #            error_file = _filename_on_where_error_raised()
    #        assert ex != None
    #        assert error_file == expected
    #        # read from cache
    #        engine.cache.clear()
    #        engine = tenjin.Engine()
    #        t = engine.get_template(fname)
    #        assert t.filename == expected
    #    def _filename_on_where_error_raised():
    #        import traceback
    #        tb = sys.exc_info()[2]
    #        arr = traceback.format_tb(tb)
    #        msg = arr[-1]
    #        m = re.search(r'File "(.*?)"', msg)
    #        return m.group(1)
    #    #
    #    if "Engine.prefer_fullpath is False then set relative epath as template filename":
    #        assert tenjin.Engine.prefer_fullpath == False
    #        try:
    #            _test(fname, input, expected=fname)
    #        finally:
    #            _remove_files([fname, fname + '.cache'])
    #    if "Engine.prefer_fullpath is True then set fullpath as template filename":
    #        tenjin.Engine.prefer_fullpath = True
    #        try:
    #            _test(fname, input, expected=os.path.join(os.getcwd(), fname))
    #        finally:
    #            tenjin.Engine.prefer_fullpath = False
    #            _remove_files([fname, fname + '.cache'])


    ##################

    def test_cachename(self):
        engine = tenjin.Engine()
        # return cache file path
        assert engine.cachename('foo.pyhtml') == 'foo.pyhtml.cache'

    def test_to_filename(self):
        engine = tenjin.Engine(prefix='user_', postfix='.pyhtml')
        # if template_name starts with ':', add prefix and postfix to it.
        assert engine.to_filename(':list') == 'user_list.pyhtml'
        # if template_name doesn't start with ':', just return it.
        assert engine.to_filename('list') == 'list'

    @_with_dummy_files
    def test__create_template(self):
        e1 = tenjin.Engine(path=['_views/blog', '_views'])
        t = None
        # if input is not specified then just create empty template object.
        t = e1._create_template(None)
        assert isinstance(t, tenjin.Template)
        assert t.filename == None
        assert t.script == None
        # if input is specified then create template object and return it.
        t = e1._create_template('<p>#{_content}</p>', '_views/layout.pyhtml')
        assert isinstance(t, tenjin.Template)
        assert t.filename == "_views/layout.pyhtml"
        assert t.script == lvars + "_extend(('''<p>''', _to_str(_content), '''</p>''', ));"

    @_with_dummy_files
    def test__preprocess(self):
        e1 = tenjin.Engine(preprocess=True)
        # preprocess template and return result
        fpath = '_views/index.pyhtml'
        input, mtime = e1.loader.load(fpath)
        ret = e1._preprocess(input, fpath, {}, globals())
        assert ret == "<<SOS>>"

    @_with_dummy_files
    def test__get_template_from_cache(self):
        e1 = tenjin.Engine(path=['_views/blog', '_views'], postfix='.pyhtml')
        fpath = '_views/blog/index.pyhtml'
        cpath = fpath + '.cache'
        # if template not found in cache, return None
        ret = e1._get_template_from_cache(cpath, fpath)
        assert ret == None
        t = tenjin.Template(fpath)
        e1.cache.set(fpath + '.cache', t)
        # if checked within a sec, skip timestamp check.
        #try:
        #    t.timestamp = time.time() - 0.2
        #    #import pdb; pdb.set_trace()
        #    tenjin.logger = DebugLogger()
        #    ret = e1._get_template_from_cache(cpath, fpath)
        #    msg = tenjin.logger.messages[0]
        #    assert msg.startswith("[TRACE] [tenjin.Engine] timestamp check skipped (") == True
        #finally:
        #    tenjin.logger = None
        pass
        # if timestamp of template objectis same as file, return it.
        t._last_checked_at = None
        t.timestamp = os.path.getmtime(fpath)
        assert e1._get_template_from_cache(cpath, fpath) is t
        delta = JYTHON and 0.03 or 0.001
        assert abs(t._last_checked_at - time.time()) <= delta
        # if timestamp of template object is different from file, clear it
        t.timestamp = t.timestamp + 1
        t._last_checked_at = None
        try:
            #import pdb; pdb.set_trace()
            tenjin.logger = DebugLogger()
            ret = e1._get_template_from_cache(cpath, fpath)
            msg = tenjin.logger.messages[0]
            assert msg == "[INFO] [tenjin.Engine] cache expired (filepath='_views/blog/index.pyhtml')"
        finally:
            tenjin.logger = None

    @_with_dummy_files
    def test_get_template(self):
        e1 = tenjin.Engine(path=['_views/blog', '_views'], postfix='.pyhtml')
        filepath  = '_views/blog/index.pyhtml'
        fullpath  = os.getcwd() + filepath
        cachepath = fullpath + '.cache'
        assert not os.path.exists(cachepath)
        t = None
        # return template object.
        # accept template_name such as ':index'.
        t = e1.get_template(':index')
        assert isinstance(t, tenjin.Template)
        assert t.filename == filepath
        # if template object is added by add_template(), return it.
        tmp = tenjin.Template('foo.pyhtml', input="<<dummy>>")
        e1.add_template(tmp)
        assert e1.get_template('foo.pyhtml') is tmp
        # get filepath and fullpath of template
        e1._filepaths['index.pyhtml'] == (filepath, fullpath)
        # if template file is not found then raise TemplateNotFoundError
        def f(): e1.get_template('index')
        with pytest.raises(tenjin.TemplateNotFoundError, match=_re.escape("index: filename not found (path=['_views/blog', '_views']).")): f()
        # use full path as base of cache file path
        # get template object from cache
        assert list(e1.cache.items.keys()) == ["%s/_views/blog/index.pyhtml.cache" % os.getcwd()]
        # change template filename according to prefer_fullpath
        pass
        # use full path as base of cache file path
        pass
        # get template object from cache
        pass
        # if template object is not found in cache or is expired...
        e1.cache.clear()
        assert len(e1.cache.items) == 0
        tname = ':layout'
        fpath = '_views/layout.pyhtml'
        cpath = os.path.join(os.getcwd(), '_views/layout.pyhtml.cache')
        assert not os.path.exists(cpath)
        # create template object.
        t = e1.get_template(tname)
        assert isinstance(t, tenjin.Template)
        # set timestamp and filename of template object.
        assert t.timestamp == os.path.getmtime(filepath)
        assert t.filename == fpath
        delta = JYTHON and 0.03 or 0.003
        assert abs(t._last_checked_at - time.time()) <= delta
        # save template object into cache.
        assert os.path.exists(cpath)
        assert len(e1.cache.items) == 1
        assert cpath in e1.cache.items

    def test_include(self):
        # get local and global vars of caller.
        pass
        # get _context from caller's local vars.
        pass
        # if kwargs specified then add them into context.
        pass
        # get template object with context data and global vars.
        pass
        # if append_to_buf is true then add output to _buf.
        pass
        # if append_to_buf is false then don't add output to _buf.
        pass
        # render template and return output.
        pass
        # kwargs are removed from context data.
        pass

    def test_hook_context(self):
        e = tenjin.Engine()
        ctx = {}
        e.hook_context(ctx)
        # add engine itself into context data.
        assert ctx.get('_engine') is e
        # add include() method into context data.
        assert ctx.get('include') == (e.include)

    def test_get_template_ignores_syntax_error_when_compiling(self):
        # get_template(): ignores syntax error when compiling.
        input = """<p>${{foo}</p>"""
        fname = "tmp_999.pyhtml"
        f = open(fname, 'w'); f.write(input); f.close()
        try:
            e = tenjin.Engine()
            def fn(): e.get_template(fname)
            fn()
        finally:
            for x in glob(fname + '*'): os.unlink(x)


_DUMMY_VALUE = 'SOS'



import tenjin
from tenjin.helpers import *
from tenjin.html import text2html
engine = tenjin.Engine(path=['views'])
context = {
    'title': 'Blog Post Test',
    'post_content': "Foo\nBar\nBaz",
}
html = engine.render('blog_post.pyhtml', context)
print(html)



import tenjin
template = tenjin.Template('views/page.pyhtml')
print(template.script)

### or:
#template = tenjin.Template()
#with open('views/page.pyhtml') as f:
#    print(template.convert(f.read(), 'views/page.pyhtml'))

### or:
#engine = tenjin.Engine(path=['views'])
#print(engine.get_template('page.pyhtml').script)



import tenjin
from tenjin.helpers import *
pp = [ tenjin.JavaScriptPreprocessor() ]
## or pp = [ tenjin.JavaScriptPreprocessor(type='text/javascript') ]
engine = tenjin.Engine(pp=pp)
context = { 'items': ["Haruhi", "Mikuru", "Yuki"] }
html = engine.render('example.pyhtml', context)
print(html)



## create Engine object
import tenjin
from tenjin.helpers import *
engine = tenjin.Engine()

## render template with context data
context = { 'title': 'Bordered Table Example',
            'items': [ '<AAA>', 'B&B', '"CCC"' ] }
output = engine.render('table.pyhtml', context)
print(output)



import tenjin
from tenjin.helpers import *

## set logging object
import logging
logging.basicConfig(level=logging.INFO)
tenjin.logger = logging

engine = tenjin.Engine()
context = {'name': 'World'}
html = engine.render('example.pyhtml', context)
#print(html)



## create Engine object
import tenjin
from tenjin.helpers import *
engine = tenjin.Engine(postfix='.pyhtml', layout='layout.pyhtml')

## render template with context data
params = { 'name': 'Foo', 'gender': 'M' }
context = { 'params': params }
output = engine.render(':update', context)   # ':update' == 'update'+postfix
print(output)
